-- (c) David Cunningham 2009, Licensed under the MIT license: http://www.opensource.org/licenses/mit-license.php

print("Loading env.lua")

local skymat = get_material("/system/SkySphere")
if sky_sphere == nil then
    sky_sphere = sm:createEntity("SkySphere","system/SkySphere.mesh")
    gfx.skyNode:attachObject(sky_sphere)
end
--sky_sphere.renderQueueGroup = 95
--skymat:setDepthBias(0,0,-26810,0)
--sky_sphere.renderQueueGroup = 5
--skymat:setDepthWriteEnabled(0,0,false)
sky_sphere.castShadows = false


function format_time (secs)
    secs = secs or 0
    return string.format("%02d:%02d:%02d", math.mod(math.floor(secs/60/60),24),
                                           math.mod(math.floor(secs/60),60),
                                           math.mod(secs,60))
end

local timer = Timer()

-- avoid GC overhead
local col_names = { "col0", "col1", "col2", "col3", "col4", "col5" }
local scol_names = { "scol0", "scol1", "scol2", "scol3", "scol4", "scol5" }

local function find_sky(secs)
    secs = secs or 0
    local current_sky, next_sky, slider
    do 
        local found = false
        for _,sky in ipairs(sky_cycle) do
            current_sky = next_sky
            next_sky = sky
            if sky.time*60*60 > secs then
                found = true
                break
            end
        end
        if not found then
            -- wrap-around case
            current_sky = sky_cycle[#sky_cycle]
            next_sky = sky_cycle[1]
            slider = (secs/60/60 - current_sky.time) / (next_sky.time+24 - current_sky.time)
        else
            slider = (secs/60/60 - current_sky.time) / (next_sky.time - current_sky.time)
        end
    end
    return current_sky, next_sky, slider
end

local function recompute_sky()

    local secs = env.secondsSinceMidnight
    -- account for the fact that the sun is on the other side of the planet in summer
    local sun_adjusted_time = math.mod(secs / 60 / 60 / 24 * 360 + env.season, 360)
    local space_orientation = Quat(env.latitude,V_EAST)*
                              Quat(sun_adjusted_time,V_SOUTH)*
                              Quat(env.earthTilt,V_WEST)
    local sun_direction =  - space_orientation * Quat(env.season,V_NORTH) * V_UP
    local moon_direction =  - space_orientation * Quat(env.moonPhase,V_NORTH) * V_UP

    -- procedural from time
    gfx.skyNode.orientation = space_orientation
    skymat:setVertexConstantFloat(0,0,"moon_pos", moon_direction.x, moon_direction.y, moon_direction.z, 0)
    skymat:setVertexConstantFloat(0,0,"sun_pos", sun_direction.x, sun_direction.y, sun_direction.z, 0)

    local current_sky, next_sky, slider = find_sky(secs)

    sun.direction = ((current_sky.moon and moon_direction or sun_direction) + 0.0*V_DOWN).normalised

    -- interpolated from sky_cycle
    sm:setAmbientLight(lerp(slider, current_sky.amb[1], next_sky.amb[1]),
                       lerp(slider, current_sky.amb[2], next_sky.amb[2]),
                       lerp(slider, current_sky.amb[3], next_sky.amb[3]))
    local fog_r, fog_g, fog_b = lerp(slider, current_sky.fog[2], next_sky.fog[2]),
                                lerp(slider, current_sky.fog[3], next_sky.fog[3]),
                                lerp(slider, current_sky.fog[4], next_sky.fog[4])
    sm:setFog("EXP2", fog_r, fog_g, fog_b, lerp(slider, current_sky.fog[1], next_sky.fog[1])/10, 0, 0)
    sun:setDiffuseColour(lerp(slider, current_sky.diff[1], next_sky.diff[1]),
                         lerp(slider, current_sky.diff[2], next_sky.diff[2]),
                         lerp(slider, current_sky.diff[3], next_sky.diff[3]))
    sun:setSpecularColour(lerp(slider, current_sky.spec[1], next_sky.spec[1]),
                          lerp(slider, current_sky.spec[2], next_sky.spec[2]),
                          lerp(slider, current_sky.spec[3], next_sky.spec[3]))
    skymat:setFragmentConstantFloat(0,0,"hell_colour", fog_r, fog_g, fog_b, 0)
    skymat:setFragmentConstantFloat(0,0,"sun_size", lerp(slider, current_sky.sunSize, next_sky.sunSize))
    skymat:setFragmentConstantFloat(0,0,"sun_falloff_distance", lerp(slider, current_sky.sunFalloffDistance, next_sky.sunFalloffDistance))
    skymat:setFragmentConstantFloat(0,0,"sun_colour", lerp(slider, current_sky.sunColour[1], next_sky.sunColour[1]),
                                                      lerp(slider, current_sky.sunColour[2], next_sky.sunColour[2]),
                                                      lerp(slider, current_sky.sunColour[3], next_sky.sunColour[3]),
                                                      lerp(slider, current_sky.sunColour[4], next_sky.sunColour[4]))
    skymat:setFragmentConstantFloat(0,0,"dividers", lerp(slider, current_sky.skyDividers[1], next_sky.skyDividers[1]),
                                                    lerp(slider, current_sky.skyDividers[2], next_sky.skyDividers[2]),
                                                    lerp(slider, current_sky.skyDividers[3], next_sky.skyDividers[3]),
                                                    lerp(slider, current_sky.skyDividers[4], next_sky.skyDividers[4]))
    skymat:setFragmentConstantFloat(0,0,"sun_glare_distance", lerp(slider, current_sky.sunGlareDistance, next_sky.sunGlareDistance))
    --skymat:setFragmentConstantFloat(0,0,"sun_glare_col0", lerp(slider, current_sky.sunGlareCol1[1], next_sky.sunGlareCol1[1]),
    --                                                      lerp(slider, current_sky.sunGlareCol1[2], next_sky.sunGlareCol1[2]),
    --                                                      lerp(slider, current_sky.sunGlareCol1[3], next_sky.sunGlareCol1[3]),
    --                                                      lerp(slider, current_sky.sunGlareCol1[4], next_sky.sunGlareCol1[4]))
    --skymat:setFragmentConstantFloat(0,0,"sun_glare_transition", lerp(slider, current_sky.sunGlareTransition, next_sky.sunGlareTransition))
    --skymat:setFragmentConstantFloat(0,0,"sun_glare_col1", lerp(slider, current_sky.sunGlareCol2[1], next_sky.sunGlareCol2[1]),
    --                                                      lerp(slider, current_sky.sunGlareCol2[2], next_sky.sunGlareCol2[2]),
    --                                                      lerp(slider, current_sky.sunGlareCol2[3], next_sky.sunGlareCol2[3]),
    --                                                      lerp(slider, current_sky.sunGlareCol2[4], next_sky.sunGlareCol2[4]))
    skymat:setFragmentConstantFloat(0,0,"horizon_glare_elevation", lerp(slider, current_sky.horizonGlareElevation, next_sky.horizonGlareElevation))
    --skymat:setFragmentConstantFloat(0,0,"horizon_glare_col0", lerp(slider, current_sky.horizonGlareCol1[1], next_sky.horizonGlareCol1[1]),
    --                                                          lerp(slider, current_sky.horizonGlareCol1[2], next_sky.horizonGlareCol1[2]),
    --                                                          lerp(slider, current_sky.horizonGlareCol1[3], next_sky.horizonGlareCol1[3]),
    --                                                          lerp(slider, current_sky.horizonGlareCol1[4], next_sky.horizonGlareCol1[4]))
    --skymat:setFragmentConstantFloat(0,0,"horizon_glare_transition", lerp(slider, current_sky.horizonGlareTransition, next_sky.horizonGlareTransition))
    --skymat:setFragmentConstantFloat(0,0,"horizon_glare_col1", lerp(slider, current_sky.horizonGlareCol2[1], next_sky.horizonGlareCol2[1]),
    --                                                          lerp(slider, current_sky.horizonGlareCol2[2], next_sky.horizonGlareCol2[2]),
    --                                                          lerp(slider, current_sky.horizonGlareCol2[3], next_sky.horizonGlareCol2[3]),
    --                                                          lerp(slider, current_sky.horizonGlareCol2[4], next_sky.horizonGlareCol2[4]))
    for i=1,6 do
        skymat:setFragmentConstantFloat(0,0,col_names[i], lerp(slider, current_sky.gradient[i][1], next_sky.gradient[i][1]),
                                                          lerp(slider, current_sky.gradient[i][2], next_sky.gradient[i][2]),
                                                          lerp(slider, current_sky.gradient[i][3], next_sky.gradient[i][3]),
                                                          lerp(slider, current_sky.gradient[i][4], next_sky.gradient[i][4]))
        skymat:setFragmentConstantFloat(0,0,scol_names[i], lerp(slider, current_sky.sunGradient[i][1], next_sky.sunGradient[i][1]),
                                                           lerp(slider, current_sky.sunGradient[i][2], next_sky.sunGradient[i][2]),
                                                           lerp(slider, current_sky.sunGradient[i][3], next_sky.sunGradient[i][3]),
                                                           lerp(slider, current_sky.sunGradient[i][4], next_sky.sunGradient[i][4]))
    end
    --sm:setSkyBox(true, "system/SkyBox", 100, true, space_orientation*Q_UP)

end

local function maybe_commit (self)
    if not self.c.autoUpdate then return end

    local reset_sun_pos = false

    for k,v in pairs(self.p) do
        if self.c[k] ~= v then
            self.c[k] = v

            if k == "secondsSinceMidnight" then
                reset_sun_pos = true
            elseif k == "latitude" then
                reset_sun_pos = true
            elseif k == "season" then
                reset_sun_pos = true
            elseif k == "earthTilt" then
                reset_sun_pos = true
            elseif k == "clockRate" then
            elseif k == "clockTicking" then
            elseif k == "moonSize" then
                skymat:setFragmentConstantFloat(0,0,"moon_size", self.moonSize)
            elseif k == "moonColour" then
                skymat:setFragmentConstantFloat(0,0,"moon_colour", unpack(self.moonColour))
            elseif k == "moonPhase" then
                reset_sun_pos = true
            else
                error("Unexpected: "..k)
            end
        end
    end

    if reset_sun_pos then
        recompute_sky()
        if env.clockTicking == false then
            update_hud()
        end
    end
end

local function change (self, k, v)
    if k == "autoUpdate" then
        ensure_one_of(v,{false,true})
        self.c[k] = v
    elseif k == "secondsSinceMidnight" then
        ensure_number(v)
        self.p[k] = v
    elseif k == "latitude" then
        ensure_range(v,-90,90)
        self.p[k] = v
    elseif k == "season" then
        ensure_range(v,0,360)
        self.p[k] = v
    elseif k == "earthTilt" then
        ensure_range(v,-90,90)
        self.p[k] = v
    elseif k == "clockRate" then
        ensure_range(v,-50000,50000)
        self.p[k] = v
    elseif k == "clockTicking" then
        ensure_one_of(v,{false,true})
        self.p[k] = v
    elseif k == "moonSize" then
        ensure_range(v,1,45)
        self.p[k] = v
    elseif k == "moonColour" then
        ensure_array(v,4)
        ensure_range(v[1],0,1)
        ensure_range(v[2],0,1)
        ensure_range(v[3],0,1)
        ensure_range(v[4],0,1)
        self.p[k] = v
    elseif k == "moonPhase" then
        ensure_range(v,0,360)
        self.p[k] = v
    else
        error("Unrecognised env setting: "..tostring(k))
    end

    maybe_commit(self)
end

function do_reset_sky_shader()
    get_gpuprog("system/SkyProgram_f"):reload()
    get_gpuprog("system/SkyProgram_v"):reload()
end


include "sky_cycle.lua"

function save_sky_cycle(filename)
    filename = filename or "saved_sky_cycle.lua"
    local f = io.open(filename,"w")
    f:write("sky_cycle = ")
    f:write(dump(sky_cycle,false))
    f:close()
end

if not env then
    env = {
        -- current values
        c = {
            autoUpdate = false;
        },

        -- proposed values
        p = {
            latitude = 34; 
            secondsSinceMidnight = 12*60*60; --midday
            season = 180; -- summer, stored as angle in degrees
            earthTilt = 23.44;
            clockRate = 60;
            clockTicking = true;
            moonPhase = 20; -- between sun and earth when moonPhase + season == 0 (mod 360)
            moonSize = 8; -- width in degrees
            moonColour = {1,1,1,1};
        }
    }

    sun = sm:createLight("Sun")
    sun.type = "DIRECTIONAL"

else

    setmetatable(env,nil)
    ui.pressCallbacks:removeByName("Environment")
    main.frameCallbacks:removeByName("Environment")
    safe_destroy(env.edittingDisplay)

end

env.edittingDisplay = get_hud_root():addChild("ShadowText",{font = "misc.fixed", charHeight = 13})
env.edittingDisplay.resize = Hud.LEFT(-13-1)
env.edittingDisplay.text = "test"
env.edittingDisplay.visible = false

env.moreRed = false
env.lessRed = false
env.moreGreen = false
env.lessGreen = false
env.moreBlue = false
env.lessBlue = false
env.moreAlpha = false
env.lessAlpha = false
env.moreSize = false
env.lessSize = false
env.edittingField = "amb"
env.edittingTime = find_sky(env.secondsSinceMidnight) 

function env:windBackwards()
    local s = self.secondsSinceMidnight/60/60
    -- search for GLB
    local current_sky  = nil
    for _,sky in ipairs(sky_cycle) do
        if sky.time < s then
            if current_sky==nil or current_sky.time<sky.time then
                current_sky = sky
            end
        end
    end
    current_sky = current_sky or sky_cycle[#sky_cycle]
    self.edittingTime = current_sky
    env.secondsSinceMidnight = current_sky.time * 60 * 60
end

function env:windForwards()
    local s = self.secondsSinceMidnight/60/60
    -- search for LUB
    local current_sky  = nil
    for _,sky in ipairs(sky_cycle) do
        if sky.time > s then
            if current_sky==nil or current_sky.time>sky.time then
                current_sky = sky
            end
        end
    end
    current_sky = current_sky or sky_cycle[1]
    self.edittingTime = current_sky
    self.secondsSinceMidnight = current_sky.time * 60 * 60
end

function env:toggleEditMode()
    if self.edittingDisplay.visible then
        self.edittingDisplay.visible = false
    else
        self:windForwards()
        self:windBackwards()
        self.edittingDisplay.visible = true
    end
end

local prev_edit = { }

local next_edit = {
    amb = "diff";
    diff = "spec";
    spec = "fog";
    fog = "grad6";
    grad6 = "grad5";
    grad5 = "grad4";
    grad4 = "grad3";
    grad3 = "grad2";
    grad2 = "grad1";
    grad1 = "sunGrad6";
    sunGrad6 = "sunGrad5";
    sunGrad5 = "sunGrad4";
    sunGrad4 = "sunGrad3";
    sunGrad3 = "sunGrad2";
    sunGrad2 = "sunGrad1";
    sunGrad1 = "sunSize";
    sunSize = "sunColour";
    sunColour = "horizonGlareLevel";
    --sunGlareCol2 = "sunGlareCol1";
    --sunGlareCol1 = "horizonGlareCol2";
    --horizonGlareCol2 = "horizonGlareCol1";
    --horizonGlareCol1 = "amb";
    horizonGlareLevel = "sunGlareRadius";
    sunGlareRadius = "amb";
}

for k,v in pairs(next_edit) do
    prev_edit[v] = k
end

local edit_details = {
    amb                   = { function(sky) return sky.amb end, 1, 2, 3, nil, nil, nil };
    diff                  = { function(sky) return sky.diff end, 1, 2, 3, nil, nil, nil };
    spec                  = { function(sky) return sky.spec end, 1, 2, 3, nil, nil, nil };
    fog                   = { function(sky) return sky.fog end, 2, 3, 4, 1, nil, nil, nil };
    grad6                 = { function(sky) return sky.gradient[6] end, 1, 2, 3, 4, nil, nil };
    grad5                 = { function(sky) return sky.gradient[5] end, 1, 2, 3, 4, function(sky) return sky.skyDividers end, 4 };
    grad4                 = { function(sky) return sky.gradient[4] end, 1, 2, 3, 4, function(sky) return sky.skyDividers end, 3 };
    grad3                 = { function(sky) return sky.gradient[3] end, 1, 2, 3, 4, function(sky) return sky.skyDividers end, 2 };
    grad2                 = { function(sky) return sky.gradient[2] end, 1, 2, 3, 4, function(sky) return sky.skyDividers end, 1 };
    grad1                 = { function(sky) return sky.gradient[1] end, 1, 2, 3, 4, nil, nil };
    sunGrad6              = { function(sky) return sky.sunGradient[6] end, 1, 2, 3, 4, nil, nil };
    sunGrad5              = { function(sky) return sky.sunGradient[5] end, 1, 2, 3, 4, function(sky) return sky.skyDividers end, 4 };
    sunGrad4              = { function(sky) return sky.sunGradient[4] end, 1, 2, 3, 4, function(sky) return sky.skyDividers end, 3 };
    sunGrad3              = { function(sky) return sky.sunGradient[3] end, 1, 2, 3, 4, function(sky) return sky.skyDividers end, 2 };
    sunGrad2              = { function(sky) return sky.sunGradient[2] end, 1, 2, 3, 4, function(sky) return sky.skyDividers end, 1 };
    sunGrad1              = { function(sky) return sky.sunGradient[1] end, 1, 2, 3, 4, nil, nil };
    sunSize               = { nil, nil, nil, nil, nil, function(sky) return sky end, "sunSize" };
    sunColour             = { function(sky) return sky.sunColour end, 1, 2, 3, 4, function(sky) return sky end, "sunFalloffDistance"};
    --sunGlareCol2          = { function(sky) return sky.sunGlareCol2 end, 1, 2, 3, 4, function(sky) return sky end, "sunGlareDistance"};
    --sunGlareCol1          = { function(sky) return sky.sunGlareCol1 end, 1, 2, 3, 4, function(sky) return sky end, "sunGlareTransition"};
    --horizonGlareCol2      = { function(sky) return sky.horizonGlareCol2 end, 1, 2, 3, 4, function(sky) return sky end, "horizonGlareElevation"};
    --horizonGlareCol1      = { function(sky) return sky.horizonGlareCol1 end, 1, 2, 3, 4, function(sky) return sky end, "horizonGlareTransition"};
    horizonGlareLevel     = { nil, nil, nil, nil, nil, function(sky) return sky end, "horizonGlareElevation" };
    sunGlareRadius        = { nil, nil, nil, nil, nil, function(sky) return sky end, "sunGlareDistance" };
}

local function append_edit(str,colours,extra_name,extra_format,extra)
    if env.edittingField == str then
        env.edittingDisplay:setColourTop(1,1,0,1)
        env.edittingDisplay:setColourBottom(1,0,0,1)
    end

    env.edittingDisplay:append(str)
    env.edittingDisplay:setColourTop(1,1,1,1)
    env.edittingDisplay:setColourBottom(1,1,1,1)

    env.edittingDisplay:append(" = ")

    if colours then
        env.edittingDisplay:append("RGB(")
        for i,col in ipairs(colours) do
            if i>1 then env.edittingDisplay:append(",") end
            env.edittingDisplay:append(string.format("%0.3f",col))
        end
        env.edittingDisplay:append(")")
    end

    if extra_name then
        if colours then env.edittingDisplay:append(" ") end
        env.edittingDisplay:append((extra_name~="" and extra_name.." " or "")..string.format(extra_format,extra))
    end

    env.edittingDisplay:append("\n")

end

function update_hud()
    env.edittingDisplay:reset()
    local sky = env.edittingTime
    append_edit("amb", sky.amb)
    append_edit("diff", sky.diff)
    append_edit("spec", sky.spec)
    append_edit("fog", {sky.fog[2], sky.fog[3], sky.fog[4]}, "strength", "%0.2f", sky.fog[1])
    append_edit("grad6", sky.gradient[6], "height", "%0.1f", 90)
    append_edit("grad5", sky.gradient[5], "height", "%0.1f", sky.skyDividers[4])
    append_edit("grad4", sky.gradient[4], "height", "%0.1f", sky.skyDividers[3])
    append_edit("grad3", sky.gradient[3], "height", "%0.1f", sky.skyDividers[2])
    append_edit("grad2", sky.gradient[2], "height", "%0.1f", sky.skyDividers[1])
    append_edit("grad1", sky.gradient[1], "height", "%0.1f", 0)
    append_edit("sunGrad6", sky.sunGradient[6], "height", "%0.1f", 90)
    append_edit("sunGrad5", sky.sunGradient[5], "height", "%0.1f", sky.skyDividers[4])
    append_edit("sunGrad4", sky.sunGradient[4], "height", "%0.1f", sky.skyDividers[3])
    append_edit("sunGrad3", sky.sunGradient[3], "height", "%0.1f", sky.skyDividers[2])
    append_edit("sunGrad2", sky.sunGradient[2], "height", "%0.1f", sky.skyDividers[1])
    append_edit("sunGrad1", sky.sunGradient[1], "height", "%0.1f", 0)
    append_edit("sunSize", nil, "", "%0.2f", sky.sunSize)
    append_edit("sunColour", sky.sunColour, "falloff", "%0.3f", sky.sunFalloffDistance)
    append_edit("horizonGlareLevel", nil, "", "%0.3f", sky.horizonGlareElevation)
    append_edit("sunGlareRadius", nil, "", "%0.3f", sky.sunGlareDistance)
    --append_edit("sunGlareCol2", sky.sunGlareCol2, "radius", "%0.3f", sky.sunGlareDistance)
    --append_edit("sunGlareCol1", sky.sunGlareCol1, "transition", "%0.3f", sky.sunGlareTransition)
    --append_edit("horizonGlareCol2", sky.horizonGlareCol2, "height", "%0.3f", sky.horizonGlareElevation)
    --append_edit("horizonGlareCol1", sky.horizonGlareCol1, "transition", "%0.3f", sky.horizonGlareTransition)
    env.edittingDisplay:append((env.edittingTime.moon and "Moon" or "Sun").." casts shadows\n\n")
    env.edittingDisplay:commit()
    env.edittingDisplay:triggerResize()
end

update_hud()

-- put behind the console so that the console still works when we're in edit mode
ui.pressCallbacks:insert("Environment",function (key)
    if not env.edittingDisplay.visible then return end
    if key=='+Left' then
        env:windBackwards()
    elseif key=='+Right' then
        env:windForwards()
    elseif key=='+Up' then
        env.edittingField = prev_edit[env.edittingField]
        update_hud()
    elseif key=='+Down' then
        env.edittingField = next_edit[env.edittingField]
        update_hud()
    elseif key=='+NUMPAD5' then
        env.edittingTime.moon = not env.edittingTime.moon
        update_hud()
    elseif key=='+NUMPAD7' then env.lessRed = true
    elseif key=='-NUMPAD7' then env.lessRed = false
    elseif key=='+NUMPAD9' then env.moreRed = true
    elseif key=='-NUMPAD9' then env.moreRed = false
    elseif key=='+NUMPAD4' then env.lessGreen = true
    elseif key=='-NUMPAD4' then env.lessGreen = false
    elseif key=='+NUMPAD6' then env.moreGreen = true
    elseif key=='-NUMPAD6' then env.moreGreen = false
    elseif key=='+NUMPAD1' then env.lessBlue = true
    elseif key=='-NUMPAD1' then env.lessBlue = false
    elseif key=='+NUMPAD3' then env.moreBlue = true
    elseif key=='-NUMPAD3' then env.moreBlue = false
    elseif key=='+NUMPAD0' then env.lessAlpha = true
    elseif key=='-NUMPAD0' then env.lessAlpha = false
    elseif key=='+NUMPAD.' then env.moreAlpha = true
    elseif key=='-NUMPAD.' then env.moreAlpha = false
    elseif key=='+NUMPAD8' then env.moreSize = true
    elseif key=='-NUMPAD8' then env.moreSize = false
    elseif key=='+NUMPAD2' then env.lessSize = true
    elseif key=='-NUMPAD2' then env.lessSize = false
    else
        return
    end
    return false
end, ui.pressCallbacks:getIndexSafe("grit_console")+1)


local function frameCallback()
    local clock_rate = env.clockRate

    local elapsed = timer.s 
    timer:reset()

    if env.clockTicking and not env.edittingDisplay.visible then

        env.secondsSinceMidnight = math.mod(env.secondsSinceMidnight + elapsed * env.clockRate, 24*60*60)

    elseif env.edittingDisplay.visible then

        local details = edit_details[env.edittingField]

        local qty = (ui:ctrl() and .1 or 1) * 60/256 * elapsed

        if details[1] then
            local tab = details[1](env.edittingTime)
            local r = details[2]
            local g = details[3]
            local b = details[4]
            local a = details[5]

            if r and env.moreRed   then tab[r] = clamp(tab[r] + qty, 0, 1) end
            if r and env.lessRed   then tab[r] = clamp(tab[r] - qty, 0, 1) end
            if g and env.moreGreen then tab[g] = clamp(tab[g] + qty, 0, 1) end
            if g and env.lessGreen then tab[g] = clamp(tab[g] - qty, 0, 1) end
            if b and env.moreBlue  then tab[b] = clamp(tab[b] + qty, 0, 1) end
            if b and env.lessBlue  then tab[b] = clamp(tab[b] - qty, 0, 1) end
            if a and env.moreAlpha then tab[a] = clamp(tab[a] + qty, 0, 1) end
            if a and env.lessAlpha then tab[a] = clamp(tab[a] - qty, 0, 1) end
        end

        if details[6] then
            local tab = details[6](env.edittingTime)
            local key = details[7]

            if env.moreSize   then tab[key] = tab[key] + qty*10 end
            if env.lessSize   then tab[key] = tab[key] - qty*10 end
        end

        recompute_sky()
        update_hud()

    end
end

main.frameCallbacks:insert("Environment", frameCallback)

setmetatable(env, {
    __index = function (self, k)
        local v = self.c[k] ;
        if v == nil then
            error("No such setting: \""..k.."\"",2)
        end
        return v
    end,
    __newindex = function (self, k, v) change(self,k,v) end
})

env.autoUpdate = true
