-- (c) David Cunningham 2009, Licensed under the MIT license: http://www.opensource.org/licenses/mit-license.php

print("Loading gfx.lua")

GFX = GFX or { }

function GFX:showStats(b)
        self.luaCounter.visible = b
        self.fpsCounter.visible = b
        self.triangleBatchCounter.visible = b
        self.memCounter.visible = b
end

function GFX.new()
        local self = { }

        self.camPos = sm.root:createChild("CamPos")
        self.camYaw = self.camPos:createChild("CamYaw") 
        self.camPitch = self.camYaw:createChild("CamPitch") 
        self.camRoll = self.camPitch:createChild("CamRoll") 

        self.skyNode = self.camPos:createChild("SkyNode") 
        self.skyNode.inheritOrientation = false

        self.camChase = sm.root:createChild("CamChase")
        self.camChase.inheritOrientation = false
        self.camChaseYaw = self.camChase:createChild("CamChaseYaw")
        self.camChasePitch = self.camChaseYaw:createChild("CamChasePitch")
        self.camChaseRoll = self.camChasePitch:createChild("CamChaseRoll")
        self.camChaseBoom = self.camChaseRoll:createChild("CamChaseBoom")
        self.camChaseBoom.position = V_FORWARDS * 8

        self.queueAggressiveness = function () return 1 + 0.5*get_in_queue_size() end


        -- OVERLAY STUFF

        self.debugTexturePane = get_hud_root():addChild("Pane")
        self.debugTexturePane.material = "DebugTexture"
        self.debugTexturePane:setUV(0,0,1,1)
        self.debugTexturePane.visible = false

        self.crosshair = get_hud_root():addChild("Pane")
        self.crosshair.material = "system/CrossHair"
        self.crosshair.resize = function (pw,ph) return pw/2-2,ph/2-2,4,4 end
       
        local opt = {font = "misc.fixed", charHeight = 13}
        self.timeCounter = get_hud_root():addChild("ShadowText",opt)
        self.timeCounter.resize = Hud.RIGHT(-65-1)

        self.fpsCounter = get_hud_root():addChild("ShadowText",opt)
        self.fpsCounter.resize = Hud.RIGHT(-52-1)

        self.streamerCounter = get_hud_root():addChild("ShadowText",opt)
        self.streamerCounter.resize = Hud.RIGHT(-39-1)

        self.luaCounter = get_hud_root():addChild("ShadowText",opt)
        self.luaCounter.resize = Hud.RIGHT(-26-1)

        self.memCounter = get_hud_root():addChild("ShadowText",opt)
        self.memCounter.resize = Hud.RIGHT(-13-1)

        self.triangleBatchCounter = get_hud_root():addChild("ShadowText",opt)
        self.triangleBatchCounter.resize = Hud.RIGHT(-1)


        self.frameTimeMin = RunningMin.new(20)
        self.frameTimeAverage = RunningAverage.new(20)
        self.frameTimeMax = RunningMax.new(20)
        self.frameTimer = Timer()

        local function frameCallback()

                sm:updateSceneGraph()
                local pos = gfx.cam.derivedPosition
                streamer:centre(pos.x,pos.y,pos.z)

                local _,unacct_allocs = get_alloc_stats()
                reset_alloc_stats()
                render() -- render a frame (takes a while!)
                local count, gfx_allocs = get_alloc_stats()
                reset_alloc_stats()

                give_queue_allowance(self:queueAggressiveness())

                local last_time = self.frameTimer.us/1000000
                self.frameTimeMin:add(last_time)
                self.frameTimeAverage:add(last_time)
                self.frameTimeMax:add(last_time)
                self.frameTimer:reset()

                if self.timeCounter and self.timeCounter.visible then
                        self.timeCounter.text = "Time of day: "..format_time(env.secondsSinceMidnight)
                end
                if self.fpsCounter and self.fpsCounter.visible then
                        local min_fps = 1/nonzero(self.frameTimeMax:calc())
                        local avg_fps = 1/nonzero(self.frameTimeAverage:calc())
                        local max_fps = 1/nonzero(self.frameTimeMin:calc())
                        self.fpsCounter.text = string.format("FPS min / avg / max: %3.0f / %3.0f / %3.0f",min_fps,avg_fps,max_fps)
                        local green = clamp(min_fps/75,0,1)
                        local red = clamp((75-min_fps)/75,0,1)
                        self.fpsCounter:setColourTop(red,green,0,1)
                        self.fpsCounter:setColourBottom(red,green,0,1)
                end
                if self.streamerCounter and self.streamerCounter.visible then
                        self.streamerCounter.text = string.format(
                                "Classes / Objects / Activated: %6.0f / %6.0f / %6.0f",
                                streamer.numClasses, streamer.numObjects, streamer.numActivated)
                end
                if self.memCounter and self.memCounter.visible then
                        self.memCounter.text = string.format(
                                "Mesh | Tex | Max || QI | QO: %dMB | %dMB | %dMB || %d | %d",
                                get_mesh_usage()/1024/1024,
                                get_texture_usage()/1024/1024,
                                (get_mesh_budget()+get_texture_budget())/1024/1024,
                                get_in_queue_size(),
                                get_out_queue_size());
                end
                if self.triangleBatchCounter and self.triangleBatchCounter.visible then
                        self.triangleBatchCounter.text = string.format(
                                "Tris | Batches: %d / %d",get_main_win().triangles, get_main_win().batches)
                end
                if self.luaCounter and self.luaCounter.visible then
                        self.luaCounter.text = string.format(
                                "Lua: %.1fMB (%dK) (G+%d) (P+%d) (U+%d)",
                                collectgarbage("count")/1024,
                                count/1000,
                                gfx_allocs,
                                physics and physics.allocs or 0,
                                unacct_allocs)
                end
        end

        main.frameCallbacks:insert("GFX.frameCallback",frameCallback)

        make_instance(self, GFX)

        self:recreateCam("MonoCamera")

        return self

end

function GFX:orient(dir) -- direction as vector
        local bearing = dir * Vector3(1,1,0) -- eliminate z component
        self.camYaw.orientation = V_NORTH:getRotationTo(bearing)
        local pitch = V_NORTH:getRotationTo(self.camYaw.orientation.inverse * dir)
        self.camPitch.orientation = pitch
        self.camRoll.orientation = Q_ID
end

function GFX:warp(pos, orientation)
        if pos ~= nil then
                self.camPos.position = pos;
        end
        if orientation ~= nil then
                local dir = V_NORTH * orientation
                self:orient(dir)
        end
end

function GFX:retWarp()
        local pos = self.cam.derivedPosition
        local orientation = self.cam.derivedOrientation
        return "gfx:warp("..tostring(pos)..","..tostring(orientation)..")"
end

function GFX:yaw()
        local pointing = gfx.cam.derivedOrientation * V_FORWARDS
        pointing = pointing * Vector3(1,1,0) -- eliminate z component
        return V_NORTH:getRotationTo(pointing)
end

function GFX:yawAngle()
        local quat = gfx:yaw()
        local angle = math.deg(2*math.acos(quat.w))
        if (gfx.cam.derivedOrientation * V_FORWARDS).x < 0 then angle = 360-angle end
        return angle
end

function GFX:destroy()
        safe_destroy(self.cam) ; self.cam = nil
        get_hud_root():removeChild(self.debugTexturePane)
        safe_destroy(self.camPos) ; self.camPos = nil
        safe_destroy(self.camYaw) ; self.camYaw = nil
        safe_destroy(self.camPitch) ; self.camPitch = nil
        safe_destroy(self.camRoll) ; self.camRoll = nil

        self.skyNode = safe_destroy(self.skyNode)

        self.camChase = safe_destroy(self.camChase)
        self.camChaseYaw = safe_destroy(self.camChaseYaw)
        self.camChasePitch = safe_destroy(self.camChasePitch)
        self.camChaseRoll = safe_destroy(self.camChaseRoll)
        self.camChaseBoom = safe_destroy(self.camChaseBoom)

        safe_destroy(self.timeCounter) self.timeCounter = nil
        safe_destroy(self.fpsCounter) self.fpsCounter = nil
        safe_destroy(self.streamerCounter) self.streamerCounter = nil
        safe_destroy(self.triangleBatchCounter) self.triangleBatchCounter = nil
        safe_destroy(self.memCounter) self.memCounter = nil
        safe_destroy(self.luaCounter) self.luaCounter = nil
        main.frameCallbacks:removeByName("GFX.frameCallback")
end

function GFX:recreateCam(type)
        if self == nil then
                error("Self is nil")
        end

        if not _G[type] then
                error("No such camera: "..type)
                return
        end
        local parent = self.camRoll
        local FOV = 55
        local nearClipDistance = 0.355
        local autoAspectRatio = true
        local useRenderingDistance = false
        if self.cam then
                parent = self.cam.parentSceneNode
                FOV = self.cam.FOV
                nearClipDistance = self.cam.nearClipDistance
                autoAspectRatio = self.cam.autoAspectRatio
                useRenderingDistance = self.cam.useRenderingDistance
                self.cam:destroy()
        end
        self.cam = _G[type].new(sm,parent,get_main_win(),1,0,0,1,1)
        self.cam.FOV = FOV
        self.cam.nearClipDistance = nearClipDistance
        self.cam.autoAspectRatio = autoAspectRatio
        self.cam.useRenderingDistance = useRenderingDistance
end


if gfx ~= nil then
        gfx:destroy()
end
gfx = GFX.new()

-- vim: shiftwidth=8:tabstop=8:expandtab
