-- (c) David Cunningham 2009, Licensed under the MIT license: http://www.opensource.org/licenses/mit-license.php

local function prog(name, lang, type)
        local p = get_gpuprog(name) or make_gpuprog(name, lang, type)
        if p.language ~= lang then
                p:destroy()
                p = make_gpuprog(name, lang, type)
        end
        return p
end

local function process_tex_name (mname, tname)
        return fqn_ex(tname, mname):sub(2)
end

local function get_colour (v)
        local r,g,b,a
        if type(v)=="number" then
                r = v % 256
                v = (v - r)/256
                g = v % 256
                v = (v - g)/256
                b = v % 256
                v = (v - b)/256
                a = v % 256
                return r/255, g/255, b/255, a/255
        else
                return v[1] or r, v[2] or g, v[3] or b, v[4] or a
        end
end

local function set_texture_filtering (mat, index)
        if debug_cfg.filtering then
                if user_cfg.anisotropy == 0 then
                        mat:setTextureFiltering(0,0,index,"LINEAR","LINEAR","LINEAR")
                else
                        mat:setTextureAnisotropy(0,0,index, user_cfg.anisotropy)
                        mat:setTextureFiltering(0,0,index,"ANISOTROPIC","ANISOTROPIC","LINEAR")
                end
        else
                mat:setTextureFiltering(0,0,index,"POINT","POINT","NONE")
        end
        mat:setMipmapBias(0,0,index,user_cfg.textureShrink);
end

local function shader_names_aux(prefix, dm, nm, spec_mode, blend, stipple, blended_bones)
        return string.format("%s:%s_%s_%s_%s_%01d_%01d",
                prefix,
                dm and "D" or "d",
                nm and "N" or "n",
                spec_mode==2 and "D" or spec_mode==1 and "S" or "s",
                stipple and "T" or "t",
                blend,
                blended_bones
        )
end

local function shader_caster_names (...)
        return shader_names_aux("cast_f", ...), shader_names_aux("cast_v", ...)
end

local function shader_names (...)
        return shader_names_aux("recv_f", ...), shader_names_aux("recv_v", ...)
end

local function make_program_caster_cg (diffuse_map,blended_bones)
        local fname, vname = shader_caster_names(diffuse_map, false, 0, 0, true, blended_bones)
        local defines = "-O3 -DUSE_DIFFUSE_MAP="..(diffuse_map and "1" or "0")
                      .." -DBLENDED_BONES="..blended_bones
                      .." -DUSE_STIPPLE_TEXTURE=1 -DNUM_SHADOW_MAPS=0"
        local tunit = 0
        if true then
                defines = defines .. " -DTEXUNIT_STIPPLE=s"..tunit
                tunit = tunit+1
        end
        if diffuse_map then
                defines = defines .. " -DTEXUNIT_DIFFUSE=s"..tunit
                tunit = tunit+1
        end

        local vp = prog(vname,"cg","VERTEX")
        vp.profiles = {"vs_3_0", "vp40"}
        vp.sourceFile = "system/uber.cg"
        vp.entryPoint = "caster_vp_main"
        vp.compileArguments = defines
        vp:setAutoConstantInt("view_proj","VIEWPROJ_MATRIX")
        vp:setAutoConstantInt("light_pos_ws","LIGHT_POSITION")
        if blended_bones > 0 then
                vp:setAutoConstantInt("bone_matrixes","WORLD_MATRIX_ARRAY_3x4")
                vp.skeletalAnimationIncluded = true
        else
                vp:setAutoConstantInt("world","WORLD_MATRIX")
        end
        vp:reload()

        local fp = prog(fname,"cg","FRAGMENT")
        fp.profiles = {"ps_3_0", "fp40"}
        fp.sourceFile = "system/uber.cg"
        fp.entryPoint = "caster_fp_main"
        fp.compileArguments = defines
        fp:setAutoConstantInt("visibility", "CUSTOM", 0)
        fp:setConstantFloat("alpha_rej", 0)
        fp:reload()
        return vp, fp
end

local function material_caster (name, original, tab)

        local shf, shv = shader_caster_names(tab.diffuseMap, false, 0, 0, true, tab.blendedBones)

        local mat = get_material(name)
        if mat==nil then
                mat = Material(name)
        else
                -- removing the technique here causes a crash because
                -- the best supported technique no-longer exists.
                mat:removeAllPasses(0)
                mat:createPass(0)
        end

        mat:setFragmentProgram(0,0,shf)
        mat:setVertexProgram(0,0,shv)

        mat:setFragmentConstantFloat(0,0,"alpha_rej", tab.alphaReject or 63/255)

        if tab.backfaces then
                mat:setCull(0,0,false)
                mat:setManualCull(0,0,false)
        end     

        local tex_index = 0

        -- stipple texture
        if true then
                mat:createTextureUnitState(0,0)
                mat:setTextureName(0,0,tex_index, "system/stipple.bmp")
                mat:setTextureFiltering(0,0,tex_index,"POINT","POINT","NONE")
                tex_index = tex_index + 1
        end

        -- diffuse map
        if tab.diffuseMap then
                mat:createTextureUnitState(0,0)
                mat:setTextureName(0,0,tex_index, process_tex_name(original,tab.diffuseMap))
                mat:setMipmapBias(0,0,tex_index,user_cfg.textureShrink);
                tex_index = tex_index + 1
        end

        return mat

end



local function make_program_cg (diffuse_map, normal_map, spec_mode, blend, stipple, blended_bones)
        local fname, vname = shader_names(diffuse_map, normal_map, spec_mode, blend, stipple, blended_bones)
        local defines = "-O3 -DUSE_VERTEX_COLOURS="..(debug_cfg.vertexColours and "1" or "0")
                      .." -DUSE_DIFFUSE_MAP="..(diffuse_map and "1" or "0")
                      .." -DUSE_NORMAL_MAP="..(normal_map and "1" or "0")
                      .." -DUSE_SPECULAR_MAP="..(spec_mode==1 and "1" or "0")
                      .." -DUSE_SPECULAR_FROM_DIFFUSE="..(spec_mode==2 and "1" or "0")
                      .." -DUSE_FOG="..(debug_cfg.fog and "1" or "0")
                      .." -DNO_VERTEX_PROCESSING="..(debug_cfg.noVertexProcessing and "1" or "0")
                      .." -DNO_FRAGMENT_PROCESSING="..(debug_cfg.noFragmentProcessing and "1" or "0")
                      .." -DUSE_STIPPLE_TEXTURE="..(stipple and "1" or "0")
                      .." -DSHADOW_RES="..user_cfg.shadowRes
                      .." -DBLEND="..tostring(blend)
                      .." -DBLENDED_BONES="..blended_bones

        if user_cfg.shadow and debug_cfg.shadowReceive then
                defines = defines .. " -DNUM_SHADOW_MAPS="..user_cfg.shadowPCSSSplits
                                  .. " -DSHADOW_PADDING="..user_cfg.shadowPCSSPadding
                                  .. " -DSHADOW_DIST1="..user_cfg.shadowPCSSEnd0
                                  .. " -DSHADOW_DIST2="..user_cfg.shadowPCSSEnd1
                                  .. " -DSHADOW_DIST3="..user_cfg.shadowPCSSEnd2
                                  .. " -DSHADOW_FADE_START="..user_cfg.shadowFadeStart
                                  .. " -DSHADOW_FADE_END="..user_cfg["shadowPCSSEnd"..(user_cfg.shadowPCSSSplits-1)]
                                  .. " -DUSE_BILINEAR_PCF_TAP="..(user_cfg.shadowPCFBilinear and "1" or "0")
                                  .. " -DPCF_SIDE="..user_cfg.shadowPCFTaps
                                  .. " -DFAKE_PCF_SIDE="..(user_cfg.shadowPCFEdgeDither and "2" or "1")
                                  .. " -DSPREAD1="..(user_cfg.shadowPCSSSpreadFactor0*user_cfg.shadowPCFSpread)
                                  .. " -DSPREAD2="..(user_cfg.shadowPCSSSpreadFactor1*user_cfg.shadowPCFSpread)
                                  .. " -DSPREAD3="..(user_cfg.shadowPCSSSpreadFactor2*user_cfg.shadowPCFSpread)
        else
                defines = defines .. " -DNUM_SHADOW_MAPS=0"
        end
        local tunit = 0
        if user_cfg.shadow and debug_cfg.shadowReceive then
                for i=1,user_cfg.shadowPCSSSplits do
                        defines = defines .. " -DTEXUNIT_SHADOW"..i.."=s"..tunit
                        tunit = tunit+1
                end
        end
        if stipple then
                defines = defines .. " -DTEXUNIT_STIPPLE=s"..tunit
                tunit = tunit+1
        end
        if diffuse_map then
                defines = defines .. " -DTEXUNIT_DIFFUSE=s"..tunit
                tunit = tunit+1
        end
        if normal_map then
                defines = defines .. " -DTEXUNIT_NORMAL=s"..tunit
                tunit = tunit+1
        end
        if spec_mode == 1 then
                defines = defines .. " -DTEXUNIT_SPECULAR=s"..tunit
                tunit = tunit+1
        end


        local vp = prog(vname,"cg","VERTEX")
        vp.profiles = {"vs_3_0", "vp40"}
        vp.sourceFile = "system/uber.cg"
        vp.compileArguments = defines
        vp.entryPoint = "receiver_vp_main"
        if not debug_cfg.noVertexProcessing then
                if blended_bones > 0 then
                        vp:setAutoConstantInt("bone_matrixes","WORLD_MATRIX_ARRAY_3x4")
                        vp.skeletalAnimationIncluded = true
                else
                        vp:setAutoConstantInt("world","WORLD_MATRIX")
                end
                vp:setAutoConstantInt("view_proj","VIEWPROJ_MATRIX")
                if user_cfg.shadow and debug_cfg.shadowReceive then
                        for i=1,user_cfg.shadowPCSSSplits do
                                vp:setAutoConstantInt("shadow_view_proj"..i,"TEXTURE_VIEWPROJ_MATRIX",i-1)
                        end
                end
                vp:setAutoConstantInt("camera_pos_ws","CAMERA_POSITION")
                vp:setAutoConstantInt("light_pos_ws","LIGHT_POSITION")
                vp:setAutoConstantInt("surf_ambient","SURFACE_AMBIENT_COLOUR")
                vp:setAutoConstantInt("surf_emissive","SURFACE_EMISSIVE_COLOUR")
                vp:setAutoConstantInt("light_ambient","AMBIENT_LIGHT_COLOUR")
        end
        vp:reload()

        local fp = prog(fname,"cg","FRAGMENT")
        fp.profiles = {"ps_3_0", "fp40"}
        fp.sourceFile = "system/uber.cg"
        fp.compileArguments = defines
        fp.entryPoint = "receiver_fp_main"
        if not debug_cfg.noFragmentProcessing then
                if debug_cfg.fog then
                        fp:setAutoConstantInt("the_fog_params", "FOG_PARAMS")
                        fp:setAutoConstantInt("the_fog_colour", "FOG_COLOUR")
                end
                fp:setAutoConstantInt("visibility", "CUSTOM", 0)
                fp:setConstantFloat("alpha_rej", 0)
                if user_cfg.shadow and debug_cfg.shadowReceive then
                        for i=1,user_cfg.shadowPCSSSplits do
                                fp:setConstantFloat("shadow_bias"..i, 0)
                        end
                end
                if diffuse_map and spec_mode == 2 then
                        fp:setConstantFloat("spec_diff_brightness", 0)
                        fp:setConstantFloat("spec_diff_contrast", 1)
                end
                if blend==4 and (diffuse_map or spec_mode == 1 or normal_map) then
                        fp:setConstantFloat("tex_size", 1024)
                end
                fp:setAutoConstantInt("light_diffuse", "DERIVED_LIGHT_DIFFUSE_COLOUR", 0)
                fp:setAutoConstantInt("light_specular", "DERIVED_LIGHT_SPECULAR_COLOUR", 0)
                fp:setAutoConstantInt("surf_diffuse", "SURFACE_DIFFUSE_COLOUR")
                fp:setAutoConstantInt("surf_specular", "SURFACE_SPECULAR_COLOUR")
                fp:setAutoConstantInt("surf_shininess", "SURFACE_SHININESS")
        end
        fp:reload()


        return vp, fp
end

local function material_ff_internal (name, original, shadow_mat, tab)

        local mat = get_material(name)
        if mat==nil then
                mat = Material(name)
        else
                mat:removeAllTechniques()
                mat:createTechnique()
                mat:createPass(0)
        end

        mat:setAmbientVertex(0,0,true)

        if tab.backfaces then
                mat:setCull(0,0,false)
                mat:setManualCull(0,0,false)
        end     

        local tex_index = 0

        -- diffuse map
        if tab.diffuseMap then
                mat:createTextureUnitState(0,0)
                mat:setTextureName(0,0,tex_index, process_tex_name(original,tab.diffuseMap))
                set_texture_filtering(mat, tex_index)
                tex_index = tex_index + 1
        end

        -- diffuse colour
        local r,g,b = get_colour(tab.diffuseColour)
        mat:setDiffuse(0,0, r,g,b,1)

        r,g,b = get_colour(tab.ambientColour,r,g,b)
        mat:setAmbient(0,0, r,g,b)

        r,g,b = get_colour(tab.specularColour)
        mat:setSpecular(0,0, r,g,b,1,tab.specularity)

        mat:setAlphaRejection(0,0, tab.alphaReject)

        -- alpha stuff
        if tab.alpha then
                mat:setSceneBlending(0,0,"SRC_ALPHA","ONE_MINUS_SRC_ALPHA")
        end
        if tab.alpha then
                mat:setDepthWriteEnabled(0,0,false)
        end
        if tab.zsort then
                mat:setTransparentSortingForced(0,0,true)
        end

        return mat
end

local function material_internal (name, original, shadow_mat, fade, tab)

        local spec_mode = 0
        if tab.specularMap then spec_mode = 1 end
        if tab.specularFromDiffuse then spec_mode = 2 end
        local shf, shv = shader_names(tab.diffuseMap, tab.normalMap, spec_mode, tab.blend, tab.stipple~=false, tab.blendedBones)

        local mat = get_material(name)
        if mat==nil then
                mat = Material(name)
        else
                mat:removeAllTechniques()
                mat:createTechnique()
                mat:createPass(0)
        end

        mat:setAmbientVertex(0,0,true)
        --mat:setDiffuseVertex(0,0,true)
        --mat:setSpecularVertex(0,0,true)
        --mat:setEmissiveVertex(0,0,true)


        mat:setShadowCasterMaterial(0,shadow_mat)

        mat:setFragmentProgram(0,0,shf)
        mat:setVertexProgram(0,0,shv)

        if not debug_cfg.noFragmentProcessing then
                mat:setFragmentConstantFloat(0,0,"alpha_rej", tab.alphaReject)
                if tab.blend > 0 then
                        mat:setFragmentConstantFloat(0,0,"tex_size", tab.blend)
                end
        end

        if tab.backfaces then
                mat:setCull(0,0,false)
                mat:setManualCull(0,0,false)
        end     

        if user_cfg.shadow and debug_cfg.shadowReceive then
                for i=1,user_cfg.shadowPCSSSplits do
                        mat:setFragmentConstantFloat(0,0,"shadow_bias"..i, tab.shadowBias)
                end
        end


        local tex_index = 0

        -- shadow maps
        if user_cfg.shadow and debug_cfg.shadowReceive then
                for i=1,user_cfg.shadowPCSSSplits do
                        mat:createTextureUnitState(0,0)
                        mat:setIsShadow(0,0,tex_index, true)
                        mat:setTextureFiltering(0,0,tex_index,"POINT","POINT","NONE")
                        tex_index = tex_index + 1
                end
        end

        -- stipple texture
        if tab.stipple ~= false then
                mat:createTextureUnitState(0,0)
                mat:setTextureName(0,0,tex_index, "system/stipple.bmp")
                mat:setTextureFiltering(0,0,tex_index,"POINT","POINT","NONE")
                tex_index = tex_index + 1
        end

        -- diffuse map
        if tab.diffuseMap then
                mat:createTextureUnitState(0,0)
                mat:setTextureName(0,0,tex_index, process_tex_name(original,tab.diffuseMap))
                set_texture_filtering(mat, tex_index)
                tex_index = tex_index + 1
        end

        -- normal map
        if tab.normalMap then
                mat:createTextureUnitState(0,0)
                mat:setTextureName(0,0,tex_index, process_tex_name(original,tab.normalMap))
                set_texture_filtering(mat, tex_index)
                tex_index = tex_index + 1
        end

        -- spec map
        if tab.specularMap then
                mat:createTextureUnitState(0,0)
                mat:setTextureName(0,0,tex_index, process_tex_name(original,tab.specularMap))
                set_texture_filtering(mat, tex_index)
                tex_index = tex_index + 1
        end

        -- spec map
        if not debug_cfg.noFragmentProcessing and tab.specularFromDiffuse and tab.diffuseMap then
                local sfd = tab.specularFromDiffuse
                mat:setFragmentConstantFloat(0,0,"spec_diff_brightness", sfd[1])
                mat:setFragmentConstantFloat(0,0,"spec_diff_contrast", sfd[2])
        end

        -- diffuse colour
        local r,g,b = get_colour(tab.diffuseColour)
        mat:setDiffuse(0,0, r,g,b,1)

        r,g,b = get_colour(tab.ambientColour)
        mat:setAmbient(0,0, r,g,b)

        r,g,b = get_colour(tab.specularColour)
        mat:setSpecular(0,0, r,g,b,1,tab.specularity)

        -- alpha stuff
        if tab.alpha or fade then
                mat:setSceneBlending(0,0,"SRC_ALPHA","ONE_MINUS_SRC_ALPHA")
        end
        if tab.alpha then
                mat:setDepthWriteEnabled(0,0,false)
        end
        if tab.zsort or fade then
                mat:setTransparentSortingForced(0,0,true)
        end

        mat.transparencyCastsShadows = tab.hasCaster or tab.stipple == false

        return mat
end

local function set_material_defaults(tab)
        tab.diffuseColour = tab.diffuseColour or 0xFFFFFFFF
        tab.ambientColour = tab.ambientColour or tab.diffuseColour
        tab.specularColour = tab.specularColour or ((tab.specularMap or tab.specularFromDiffuse) and 0xFFFFFFFF or 0x00000000)
        tab.hasCaster = tab.alpha and tab.diffuseMap and user_cfg.shadowUseDiffuse
        tab.blendedBones = tab.blendedBones or 0
        tab.blend = tab.blend and tab.blend[1] or 0
        if not (tab.diffuseMap or tab.normalMap or tab.specularMap) then tab.blend = 0 end
        tab.alphaReject = tab.alphaReject or 63/255
        tab.specularity = tab.specularity or 10
        tab.shadowBias = tab.shadowBias or (tab.backfaces and 0.005 or -0.005)
        return tab
end

-- curry to avoid parentheses
local function material3(name,tab)
        if debug_cfg.diffuseMaps == false then tab.diffuseMap = nil end
        if debug_cfg.normalMaps == false then tab.normalMap = nil end
        if debug_cfg.specularMaps == false then tab.specularMap = nil end

        set_material_defaults(tab)

        local caster = tab.hasCaster and material_caster(name.."!", name, tab) or default_caster[tab.blendedBones]
        if debug_cfg.fixedFunction then
                material_ff_internal(name, name, caster, tab)
        else
                if tab.stipple == false then
                        material_internal(name.."'", name, caster, true, tab)
                end
                material_internal(name, name, caster, false, tab)
        end
end

-- use global to pass name to curried function because an upvalue implies an allocation
local the_name = nil

local function material2(tab)
        local name = fqn(the_name)
        tab = tab or {}
        --tab.stipple = false
        register_material(name,tab)
        material3(name,tab)
end

function material(name) the_name = name ; return material2 end


-- make sure global variable exists, to avoid error later one
default_caster = default_caster or {}


function do_reset_shaders ()
        local caster = make_program_caster_cg
        local receiver = make_program_cg

        for _,diff in ipairs{false,true} do
                for _,bones in ipairs{0,1,2,3,4} do
                        caster(diff,bones)
                end
        end

        local tab = set_material_defaults {}
        -- always rebuild default caster
        for _,bones in ipairs{0,1,2,3,4} do
                local name = shader_names_aux("default_caster", false, false, 0, 0, true, bones);
                tab.blendedBones = bones
                material_caster(name.."!", name, tab)
                default_caster[bones] = get_material(name.."!")
        end

        for _,diff in ipairs{false,true} do
                for _,normal in ipairs{false,true} do
                        for _,spec in ipairs{0,1,2} do
                                for _,blend in ipairs{0,4} do
                                        for _,stipple in ipairs{false,true} do
                                                for _,bones in ipairs{0,1,2,3,4} do
                                                        receiver(diff,normal,spec,blend,stipple,bones)
                                                end
                                        end
                                end
                        end
                end
        end

end

function do_reset_materials()
        --regenerate_materials()
        --BaseWhite is the only material without a leading / so we do it directly
        register_material("BaseWhite",{})
        material3("BaseWhite",{})
        --One with a leading / too
        material "/BaseWhite" { }
        reprocess_all_registered_materials(material3)
end



-- vim: ts=8:sw=8:et
